/*
 * +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * 		PROPRIETARY NOTICE (Combined)
 *
 * This source code is unpublished proprietary information
 * constituting, or derived under license from AT&T's UNIX(r) System V.
 * In addition, portions of such source code were derived from Berkeley
 * 4.3 BSD under license from the Regents of the University of
 * California.
 *
 *
 *
 * 		Copyright Notice
 *
 * Notice of copyright on this source code product does not indicate
 * publication.
 *
 * 	(c) 1991  Sun Microsystems, Inc
 *	All rights reserved.
 *
 */

/*
 * This module implements the __switch_gethostbyYY interface which
 * is implemented via the naming system back-ends in
 * the order specified by the switch configuration file, until the name is
 * resolved, a hard error encountered or it runs out of all modules.
 *
 * This interface is **internal** and used only by the switch.so
 * nametoaddr library in the /etc/netconfig file. There are no equivalent
 * __switch_gethostent routines, because the real XXXhostent routines
 * are directly implemented to follow the switch configuration policy;
 * [as opposed to the real gethostbyYY routines that call netdir_getbyYY
 * interface, which in turn may call the __switch_gethostbyYY interface].
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/file.h>
#include <dlfcn.h>
#include <limits.h>
#include <nsswitch.h>
#include <syslog.h>

static struct translator {
	struct hostent *(*hpname)();
	struct hostent *(*hpaddr)();
	char	module_name[_POSIX_PATH_MAX];
	void	*module_handle;
	struct translator	*next;
};

static struct namaddr {
	int u_type;
	union {
		char *nam;
		struct addr {
			char *adr;
			int length;
			int type;
		} addr;
	} u;
};

#define	NAME 1
#define	ADDR 2

/* default policy *not* the same as in SunOS 4.1 */
#define	DEF_ACTION {1, 0, 0, 0}
static struct __nsw_lookup lkp2 = { "nis", DEF_ACTION, NULL, NULL};
static struct __nsw_lookup lkp1 = { "files", DEF_ACTION, NULL, &lkp2};
static struct __nsw_switchconfig hosts_default = { 0, "hosts", 2, &lkp1 };

static struct translator *nsxlate_list = NULL;
static struct __nsw_switchconfig *conf = NULL;

static struct hostent *gethostbynamadr();
static struct hostent *callit();
static struct translator *load_xlate();

struct hostent *
__switch_gethostbyname(name)
	char *name;
{
	struct namaddr namaddr;

	namaddr.u_type = NAME;
	namaddr.u.nam = name;
	return (gethostbynamadr(namaddr));
}

struct hostent *
__switch_gethostbyaddr(addr, length, type)
	char *addr;
	int length;
	int type;
{
	struct namaddr namaddr;

	namaddr.u_type = ADDR;
	namaddr.u.addr.adr = addr;
	namaddr.u.addr.length = length;
	namaddr.u.addr.type = type;
	return (gethostbynamadr(namaddr));
}

static struct hostent *
gethostbynamadr(namaddr)
	struct namaddr namaddr;
{
	struct	translator   *xlate; /* pointer to translator list */
	char	*cur_module;
	struct	__nsw_lookup *lkp;
	struct	hostent *hostent = NULL;
	enum	__nsw_parse_err pserr;
	int	nserr;

	if (!conf &&
		(conf = __nsw_getconfig(__NSW_HOSTS_DB, &pserr)) == NULL)
		conf = &hosts_default;
	for (lkp = conf->lookups; lkp; lkp = lkp->next) {
		cur_module = lkp->service_name;
		nserr = __NSW_SUCCESS;
		xlate = load_xlate(&nserr, cur_module);
		if (!xlate && __NSW_ACTION(lkp, nserr) == __NSW_RETURN)
			break;
		if (!xlate)
			continue;
		nserr = __NSW_SUCCESS;
		hostent = callit(xlate, namaddr, &nserr);
		if (__NSW_ACTION(lkp, nserr) == __NSW_RETURN)
			break;
	}
	/*
	 * POLICY: This implementation is different from that for
	 * other 7 databases, in that it does not try to reset
	 * the current pointer of the corresponding getXXXent
	 * routines via a endXXXent. In other words, gethostent
	 * and gethostbyYY calls do not interact at all.
	 */
	/* XXX __nsw_freeconfig(conf); free mem */
	return (hostent);
}

static struct translator *
load_xlate(nserr, module_name)
	int *nserr;
	char *module_name;
{
	struct  translator   *xlate;
	char    full_mod_name[_POSIX_PATH_MAX];
	char    *mp = full_mod_name;
	void    *mod_handle;

	for (xlate = nsxlate_list; xlate; xlate = xlate->next)
		if (strcmp(module_name, xlate->module_name) == 0)
			return (xlate);
	memset(mp, 0, _POSIX_PATH_MAX);
	sprintf(mp, "%s%s%s", __NSW_LIB, module_name, ".so");
	mod_handle = dlopen(mp, RTLD_LAZY);
	if (mod_handle == NULL) {
		*nserr = __NSW_UNAVAIL;
#ifdef DEBUG
		syslog(LOG_ALERT, "%s\n", dlerror());
#endif
		return (NULL);
	}
	xlate = (struct translator *)malloc(sizeof (struct translator));
	if (xlate == NULL) {
		*nserr = __NSW_UNAVAIL; /* XXX: really means something else ! */
		return (NULL);
	}
	xlate->next = nsxlate_list;
	nsxlate_list = xlate;
	xlate->module_handle = mod_handle;
	strcpy(xlate->module_name, module_name);
	xlate->hpname = (struct hostent *(*)())dlsym(mod_handle,
				    "_gethostbyname");
	xlate->hpaddr = (struct hostent *(*)())dlsym(mod_handle,
				"_gethostbyaddr");
	*nserr = __NSW_SUCCESS;
	return (xlate);
}

/*
 * Do the actual work of calling the appropriate routine by
 * switching on namaddr, return the hostent if successful;
 * otherwise NULL and error.
 */

static struct hostent *
callit(xlate, namaddr, errp)
struct translator *xlate;
struct namaddr namaddr;
int *errp;
{
	struct hostent *hp = NULL;
#ifdef DEBUG
	char *dl;
#endif

	*errp = __NSW_SUCCESS;
	switch (namaddr.u_type) {
	case NAME:
#ifdef DEBUG
		syslog(LOG_DEBUG, "gethostbyname: in %s for %s\t",
			xlate->module_name, namaddr.u.nam);
#endif
		if (!(xlate->hpname))
			*errp = __NSW_UNAVAIL;
		else
			hp = (*(xlate->hpname))(errp, namaddr.u.nam);
#ifdef DEBUG
		if (hp)
			syslog(LOG_DEBUG, "succeeds\n");
		else
			syslog(LOG_DEBUG, "fails !\n");
#endif
		break;
	case ADDR:
#ifdef DEBUG
		syslog(LOG_DEBUG, "gethostbyaddr: in %s for %s\t",
			xlate->module_name, namaddr.u.addr.adr);
#endif
		if (!(xlate->hpaddr))
			*errp = __NSW_UNAVAIL;
		else
			hp = (*(xlate->hpaddr))(errp,
				namaddr.u.addr.adr,
				namaddr.u.addr.length,
				namaddr.u.addr.type);
#ifdef DEBUG
		if (hp)
			syslog(LOG_DEBUG, "succeeds\n");
		else
			syslog(LOG_DEBUG, "fails !\n");
#endif
		break;
	default:
		break;
	}
#ifdef DEBUG
	dl = dlerror();
	if (dl)
		syslog(LOG_ALERT, "gethost: backend %s\n", dl);
#endif
	return (hp);
}
